#ifndef _BCACHEFS_VENDOR_BIO_IOV_ITER_H
#define _BCACHEFS_VENDOR_BIO_IOV_ITER_H

int bch2_bio_iov_iter_get_pages(struct bio *, struct iov_iter *, unsigned);

#endif /* _BCACHEFS_VENDOR_BIO_IOV_ITER_H */
